#!/bin/sh

# Copyright (C) 2017-2020 Centrify Corporation. All rights reserved.
#
# install.sh - Centrify DirectControl openldap install script for CoreOS.
#

CIMS_USR=/opt/centrify
CIMS_INSTALL=/opt/centrify
export PATH=$CIMS_INSTALL/bin:/bin:/usr/bin:$PATH

PKG_NAME=CentrifyDC-openldap
REQUIRED_PKGS="CentrifyDC-openssl"
DEPENDENT_PKGS="CentrifyDC"
LOG_FILE=/var/log/centrifydc-install.log

PRE_EXT=.pre_cdc
TMP_EXT=.tmp
LNK_EXT=.pre_cdc_link
LST_EXT=.lst

TRUE=0
FALSE=1
DO_LOG=$FALSE
DO_FORCE=$FALSE
DO_VERBOSE=$FALSE
DO_SKIP=$FALSE
ACT=
ACTED=
ACTING=
RUN_RESULT=
UPDATE_RESULT=

PROG="`basename $0 .sh`"
PROG_FULL="$0"
COMMAND_LINE="$0 $*"
TGZ=`ls $PKG_NAME-[0-9]*-files.tgz 2> /dev/null | sort -r | head -1`
VERSION="`echo $TGZ | cut -d'-' -f3-4`"
TGZ_FILELIST_NEW="$PKG_NAME-$VERSION$LST_EXT"

TMP_FILE="/tmp/$PROG.sh.$$"
LOG_ERROR="/dev/null"

# Signals need to be caught: 1=SIGHUP 2=SIGINT 3=SIGQUIT 15=SIGTERM
SIGNAL_LIST="1 2 3 15"

usage()
{
    cat <<EOF

This script installs/uninstalls Centrify DirectControl openldap component.
Only the superuser can run this script.

Usage:
  $PROG.sh [options...]

where:
  -i            Install
  -e            Uninstall
  -u            Upgrade
  -r            Reinstall
  -f            Force operation (use with -e option)
  -s            Skip package scripts
  -p            Print verbose messages
  -h            Display this help page

EOF
}

log()
{
    if [ "$DO_LOG" = $TRUE ]; then
        echo "$*" >> "$LOG_FILE"
    fi
}

print()
{
    echo "$*"
    log "$@"
}

print2()
{
    echo "$*" >&2
    log "$@"
}

verbose()
{
    if [ "$DO_VERBOSE" = $TRUE ]; then
        print "$@"
    else
        log "$@"
    fi
}

warn()
{
    print2 "WARNING:" "$@"
}

error()
{
    print2 "ERROR:" "$@"
}

run()
{
    verbose "# $*"
    RUN_RESULT=`eval $@ 2>> "$LOG_ERROR"`
    verbose "$RESULT"
}

symlink_to_file()
{
    if [ -h "$1" ]; then
        FILE=`readlink -f "$1"`
        # Backup symlink
        test -e "$1$LNK_EXT" || \
            cp -pfd "$1" "$1$LNK_EXT" 2> /dev/null || true
        # Use mv to make the change atomic
        verbose "Replacing symlink $1 with $FILE ..."
        rm -f "$1$TMP_EXT"
        cp -pf "$FILE" "$1$TMP_EXT"
        mv -f "$1$TMP_EXT" "$1"
    fi
}

is_directory_clean()
{
    if [ -n "$1" ]; then
        verbose "Checking directory $1 ..."
        RUN_RESULT=`find $1/ ! \( -path "$1/" -o -path "$TGZ_FILELIST_INSTALLED" \) 2> /dev/null` || true
        if [ -n "$RUN_RESULT" ]; then
            verbose "File(s) left in $1:"
            for FILE in $RUN_RESULT ; do
                verbose "  $FILE"
            done
            return 1
        fi
    fi
    return 0
}

remove_directory()
{
    if [ -n "$1" ]; then
        is_directory_clean "$1"
        verbose "Removing directory $1 ..."
        rm -rf "$1"
    fi
}

get_version_number()
{
    if [ -n "$1" ]; then
        echo "$1" | sed 's/[^0-9]*//g'
    else
        echo 0
    fi
}

get_current_version()
{
    echo "$TGZ_FILELIST_INSTALLED" | sed -e "s/$LST_EXT$//" | cut -d'-' -f3-4
}

get_package_version()
{
    PKG_FILELIST="`ls -1 $CIMS_INSTALL/$1-[0-9]*$LST_EXT 2> /dev/null`"
    if test -r "$PKG_FILELIST" ; then
        basename "$PKG_FILELIST" | sed -e "s/^$1-//" -e "s/$LST_EXT$//" | cut -d'-' -f1-2
    fi
}

get_installed_dependent_packages()
{
    INSTALLED_PKGS=
    for DEPENDENT_PKG in $DEPENDENT_PKGS ; do
        DEPENDENT_VERSION="`get_package_version $DEPENDENT_PKG`"
        if test -n "$DEPENDENT_VERSION" ; then
            # Installed
            INSTALLED_PKGS="$INSTALLED_PKGS $DEPENDENT_PKG"
        fi
    done
    echo $INSTALLED_PKGS
}

are_all_required_packages_installed()
{
    ALL_INSTALLED=$TRUE
    for REQUIRED_PKG in $REQUIRED_PKGS ; do
        REQUIRED_VERSION="`get_package_version $REQUIRED_PKG`"
        if test -z "$REQUIRED_VERSION" ; then
            # Not installed
            verbose "Required package $REQUIRED_PKG is not installed."
            ALL_INSTALLED=$FALSE
        fi
    done
    return $ALL_INSTALLED
}

is_installed()
{
    TGZ_FILELIST_INSTALLED="`ls -1 $CIMS_INSTALL/$PKG_NAME-[0-9]*$LST_EXT 2> /dev/null`"
    test -r "$TGZ_FILELIST_INSTALLED"
}

install_tgz()
{
    print "Start $ACTING $PKG_NAME $VERSION ..."
    (
        set -e

        test "$DO_SKIP" = $TRUE || install_tgz_pre

        verbose "Extracting $TGZ ..."
        gunzip -c "$TGZ" | (cd / && tar xfp -) 2>> "$LOG_ERROR"

        test "$DO_SKIP" = $TRUE || install_tgz_post

        tar tf "$TGZ" | tac > "$CIMS_INSTALL/$TGZ_FILELIST_NEW" 2>> "$LOG_ERROR"
        cp -pf $PROG_FULL $CIMS_INSTALL/bin/ 2>> "$LOG_ERROR"
        echo ".$CIMS_INSTALL/bin/`basename $PROG_FULL`" >> "$CIMS_INSTALL/$TGZ_FILELIST_NEW" 2>> "$LOG_ERROR"

        if [ "$ACT" = upgrade -a "$VERSION" != "$CURRENT_VERSION" ]; then

            # Remove files and symlinks that is not in the new package
            verbose "Cleaning old files ..."
            while read -r line ; do
                if ! grep "^$line$" "$CIMS_INSTALL/$TGZ_FILELIST_NEW" > /dev/null ; then
                    if [ -f "/$line" -o -h "/$line" ]; then
                        verbose "  deleting /$line"
                        rm -f "/$line" || true
                    fi
                fi
            done < "$TGZ_FILELIST_INSTALLED" 2>> "$LOG_ERROR"

            # Remove directories that is not in the new package
            verbose "Cleaning old directories ..."
            DIR_REMOVED=$TRUE
            while test "$DIR_REMOVED" = $TRUE ; do
                DIR_REMOVED=$FALSE
                while read -r line ; do
                    if ! grep "^$line$" "$CIMS_INSTALL/$TGZ_FILELIST_NEW" > /dev/null ; then
                        if [ -d "/$line" -a "/$line" != "/./" ]; then
                            # Directory may belong to system,
                            # do not show error
                            verbose "  deleting /$line"
                            rmdir "/$line" 2> /dev/null && DIR_REMOVED=$TRUE || true
                        fi
                    fi
                done < "$TGZ_FILELIST_INSTALLED" 2>> "$LOG_ERROR"
            done

            rm "$TGZ_FILELIST_INSTALLED" 2>> "$LOG_ERROR" || true
        fi
    )

    RC=$?
    if [ "$ACT" = install -a $RC -ne 0 ]; then
        rm -f "$CIMS_INSTALL/$TGZ_FILELIST_NEW"
    fi

    return $RC
}

uninstall_tgz()
{
    print "Start $ACTING $PKG_NAME $CURRENT_VERSION ..."
    (
        set -e

        if [ ! -r "$TGZ_FILELIST_INSTALLED" ]; then
            error "Cannot read $TGZ_FILELIST_INSTALLED"
            false
        fi

        test "$DO_SKIP" = $TRUE || uninstall_tgz_pre

        # Remove files and symlinks
        verbose "Removing package files ..."
        while read -r line ; do
            if [ -f "/$line" -o -h "/$line" ]; then
                rm -f "/$line" || true
            fi
        done < "$TGZ_FILELIST_INSTALLED" 2>> "$LOG_ERROR"

        # Remove directories
        verbose "Removing package directories ..."
        DIR_REMOVED=$TRUE
        while test "$DIR_REMOVED" = $TRUE ; do
            DIR_REMOVED=$FALSE
            while read -r line ; do
                if [ -d "/$line" -a "/$line" != "/./" ]; then
                    # Directory may belong to system,
                    # do not show error
                    rmdir "/$line" 2> /dev/null && DIR_REMOVED=$TRUE || true
                fi
            done < "$TGZ_FILELIST_INSTALLED" 2>> "$LOG_ERROR"
        done

        test "$DO_SKIP" = $TRUE || uninstall_tgz_post

        rm -f "$TGZ_FILELIST_INSTALLED"
    )
    
    if [ "$DO_FORCE" = $TRUE ] && is_installed ; then
        print "Force $ACTING ..."
        rm -f "$TGZ_FILELIST_INSTALLED"
    fi
    
    test ! -f "$TGZ_FILELIST_INSTALLED"
}

install_tgz_pre()
{
    # Check if /etc/logrotate.d/centrifydc is a symlink
    LOGROTATE_CONF="/etc/logrotate.d/centrifydc"
    if [ -h "${LOGROTATE_CONF}" ]; then
        echo "ERROR: The file ${LOGROTATE_CONF} is a symbolic link which is not expected." \
            "Cannot proceed with upgrade/installation." \
            "Please contact Centrify Technical Support." | tee -a "$LOG_ERROR"
        exit 1
    fi
}

install_tgz_post()
{
    :
}

uninstall_tgz_pre()
{
    :
}

uninstall_tgz_post()
{
    :
}

clean_up()
{
    rm -f "$TMP_FILE"
}

clean_exit()
{
    clean_up
    exit $1
}

signal_handler()
{
    trap "" ${SIGNAL_LIST}
    error "Signal received."
    clean_up
    exit 1
}


# Parse arguments
while getopts :ieurfsph OPT; do
    case $OPT in
        i)
            ACT=install
            ;;
        e)
            ACT=uninstall
            ;;
        u)
            ACT=upgrade
            ;;
        r)
            ACT=reinstall
            ;;
        f)
            DO_FORCE=$TRUE
            ;;
        s)
            DO_SKIP=$TRUE
            ;;
        p)
            DO_VERBOSE=$TRUE
            ;;
        h)
            usage
            exit 0
            ;;
        \?)
            error "Invalid option '-$OPTARG'."
            exit 1
            ;;
        :)
            error "'-$OPTARG' requires an argument."
            exit 1
            ;;
    esac
done
shift $((OPTIND-1))

if [ -z "$ACT" ]; then
    error "No operation is specified."
    usage
    exit 1
fi

case "$ACT" in
    install|upgrade|reinstall|uninstall)
        if [ "`id -u`" != 0 ]; then
            error "This script must be run by root."
            exit 1
        fi
        ;;
esac

# Write a marker to the log file
DO_LOG=$TRUE
LOG_ERROR=$LOG_FILE
log ""
log "--- $PROG.sh ---:" `date`
log "Command line: $COMMAND_LINE"

trap signal_handler ${SIGNAL_LIST}
touch "$TMP_FILE" 2>> "$LOG_ERROR"

# Preprocess for install, upgrade and reinstall
case "$ACT" in
    install|upgrade|reinstall)
        if [ -f "$TGZ" ]; then
            if [ "$ACT" = upgrade -o "$ACT" = reinstall ] && ! is_installed ; then
                ACT=install
            fi
            if [ "$ACT" = install ]; then
                ACTED=installed
                ACTING=installing
                if ! is_installed ; then
                    if ! are_all_required_packages_installed ; then
                        error "Please install all required packages first!"
                        clean_exit 1
                    fi
                else
                    error "$PKG_NAME is already installed."
                    clean_exit 1                    
                fi
            else
                VERSION_NUM=`get_version_number "$VERSION"`
                CURRENT_VERSION=`get_current_version`
                CURRENT_VERSION_NUM=`get_version_number "$CURRENT_VERSION"`
                if [ -n "$CURRENT_VERSION" ]; then
                    if [ "$ACT" = upgrade ]; then
                        ACTED=upgraded
                        ACTING=upgrading
                        if [ $VERSION_NUM -lt $CURRENT_VERSION_NUM ]; then
                            error "$PKG_NAME installed version $CURRENT_VERSION is newer than $VERSION."
                            clean_exit 1
                        fi
                    else
                        ACTED=reinstalled
                        ACTING=reinstalling
                        if [ "$CURRENT_VERSION" != "$VERSION" ]; then
                            error "$PKG_NAME installed version is not $VERSION."
                            clean_exit 1
                        fi
                    fi
                else
                    error "$PKG_NAME version cannot be determined."
                    clean_exit 1
                fi
            fi
        else
            error "$PKG_NAME tarball not found."
            clean_exit 1
        fi
        ;;
esac

case "$ACT" in
    install|upgrade|reinstall)

        install_tgz
        RC=$?
        if [ $RC -eq 0 ]; then
            print "$PKG_NAME is successfully $ACTED."
        else
            error "Cannot $ACT $PKG_NAME."
            clean_exit 1
        fi
        ;;
    uninstall)
        ACTED=uninstalled
        ACTING=uninstalling
        if is_installed ; then

            CURRENT_VERSION=`get_current_version`
            CURRENT_VERSION_NUM=`get_version_number "$CURRENT_VERSION"`
            DEPENDENT_PKGS=`get_installed_dependent_packages`
            if [ -n "$DEPENDENT_PKGS" ]; then
                error "Cannot $ACT because ${DEPENDENT_PKGS} are currently installed."
                clean_exit 1
            fi

            uninstall_tgz
            RC=$?
            if [ $RC -eq 0 ]; then
                print "$PKG_NAME is successfully $ACTED."
            else
                error "Cannot $ACT $PKG_NAME."
                clean_exit 1
            fi
        else
            error "$PKG_NAME is not installed."
            clean_exit 1
        fi
        ;;
    *)
        error "Unknown operation."
        clean_exit 1
esac

clean_up
exit 0
