#!/bin/sh

# Copyright (C) 2017-2020 Centrify Corporation. All rights reserved.
#
# install.sh - Centrify DirectControl install script for CoreOS.
#

CIMS_USR=/opt/centrify
CIMS_INSTALL=/opt/centrify
export PATH=$CIMS_INSTALL/bin:/bin:/usr/bin:$PATH

PKG_NAME=CentrifyDC
REQUIRED_PKGS="CentrifyDC-openssl CentrifyDC-openldap CentrifyDC-curl"
DEPENDENT_PKGS="CentrifyDC-utest"
LOG_FILE=/var/log/centrifydc-install.log

PRE_EXT=.pre_cdc
TMP_EXT=.tmp
LNK_EXT=.pre_cdc_link
LST_EXT=.lst
INJECT_ENV="LD_LIBRARY_PATH=$CIMS_INSTALL/system/lib64"
SYSTEMD_CONF=/etc/systemd/system.conf
SYSTEMD_DIR=/etc/systemd/system
PAM_ETC_DIR=/etc/pam.d
PAM_USR_DIR=/usr/lib/pam.d
SSL_CERTS_DIR=/etc/ssl/certs

TRUE=0
FALSE=1
DO_LOG=$FALSE
DO_FORCE=$FALSE
DO_VERBOSE=$FALSE
DO_SKIP=$FALSE
ACT=
ACTED=
ACTING=
RUN_RESULT=
UPDATE_RESULT=

PROG="`basename $0 .sh`"
PROG_FULL="$0"
COMMAND_LINE="$0 $*"
TGZ=`ls $PKG_NAME-[0-9]*-files.tgz 2> /dev/null | sort -r | head -1`
VERSION="`echo $TGZ | cut -d'-' -f2-3`"
TGZ_FILELIST_NEW="$PKG_NAME-$VERSION$LST_EXT"

TMP_FILE="/tmp/$PROG.sh.$$"
LOG_ERROR="/dev/null"

# Signals need to be caught: 1=SIGHUP 2=SIGINT 3=SIGQUIT 15=SIGTERM
SIGNAL_LIST="1 2 3 15"

GET_PROCESS_PID="PROCESS_PID=\`pgrep -x \$PROCESS\` && test -n \"\$PROCESS_PID\" && ps -o pid= -o cgroup= \$PROCESS_PID | grep -v '/docker/' | grep -v  '/system.slice/docker-' | awk '{ print \$1 }'"

usage()
{
    cat <<EOF

This script installs/uninstalls Centrify DirectControl service.
Only the superuser can run this script.

Usage:
  $PROG.sh [options...]

where:
  -i            Install
  -e            Uninstall
  -u            Upgrade
  -r            Reinstall
  -f            Force operation (use with -e option)
  -s            Skip package scripts
  -p            Print verbose messages
  -h            Display this help page

EOF
}

log()
{
    if [ "$DO_LOG" = $TRUE ]; then
        echo "$*" >> "$LOG_FILE"
    fi
}

print()
{
    echo "$*"
    log "$@"
}

print2()
{
    echo "$*" >&2
    log "$@"
}

verbose()
{
    if [ "$DO_VERBOSE" = $TRUE ]; then
        print "$@"
    else
        log "$@"
    fi
}

warn()
{
    print2 "WARNING:" "$@"
}

error()
{
    print2 "ERROR:" "$@"
}

run()
{
    verbose "# $*"
    RUN_RESULT=`eval $@ 2>> "$LOG_ERROR"`
    verbose "$RESULT"
}

symlink_to_file()
{
    if [ -h "$1" ]; then
        FILE=`readlink -f "$1"`
        # Backup symlink
        test -e "$1$LNK_EXT" || \
            cp -pfd "$1" "$1$LNK_EXT" 2> /dev/null || true
        # Use mv to make the change atomic
        verbose "Replacing symlink $1 with $FILE ..."
        rm -f "$1$TMP_EXT"
        cp -pf "$FILE" "$1$TMP_EXT"
        mv -f "$1$TMP_EXT" "$1"
    fi
}

update_file()
{
    UPDATE_RESULT=
    if [ -n "$1" -a -n "$2" ]; then
        symlink_to_file "$1" || true
        if [ ! -f "$1" ]; then
            verbose "Creating $1 ..."
            touch "$1"
        fi
        CHECK_PATTERN=${3:-"^[ 	]*$2[ 	]*$"}
        if ! grep "$CHECK_PATTERN" "$1" > /dev/null ; then
            test -e "$1$PRE_EXT" || \
                cp -p "$1" "$1$PRE_EXT" || true
            verbose "Updating $1 ..."
            (
                rm -f "$1$TMP_EXT"
                cp -pf "$1" "$1$TMP_EXT"
                echo "$2" >> "$1$TMP_EXT"
                mv -f "$1$TMP_EXT" "$1"
            )
            UPDATE_RESULT=updated
        else
            UPDATE_RESULT=unchanged
        fi
    fi
}

copy_pam_file()
{
    if [ -n "$1" ]; then
        if [ ! -f "$PAM_ETC_DIR/$1" ]; then
            if [ -f "$PAM_USR_DIR/$1" ]; then
                verbose "Copying $PAM_ETC_DIR/$1 from $PAM_USR_DIR ..."
                cp -pf "$PAM_USR_DIR/$1" "$PAM_ETC_DIR/$1"
            else
                warn "Cannot find PAM configuration file $1."
            fi
        fi
    fi
}

install_systemd_service()
{
    if [ -n "$1" ]; then
        verbose "Installing $1 ..."
        cp -pf "$CIMS_INSTALL/etc/systemd/system/$1" \
            "$SYSTEMD_DIR"
    fi
}

is_directory_clean()
{
    if [ -n "$1" ]; then
        verbose "Checking directory $1 ..."
        RUN_RESULT=`find $1/ ! \( -path "$1/" -o -path "$TGZ_FILELIST_INSTALLED" \) 2> /dev/null` || true
        if [ -n "$RUN_RESULT" ]; then
            verbose "File(s) left in $1:"
            for FILE in $RUN_RESULT ; do
                verbose "  $FILE"
            done
            return 1
        fi
    fi
    return 0
}

remove_directory()
{
    if [ -n "$1" ]; then
        is_directory_clean "$1"
        verbose "Removing directory $1 ..."
        rm -rf "$1"
    fi
}

get_version_number()
{
    if [ -n "$1" ]; then
        echo "$1" | sed 's/[^0-9]*//g'
    else
        echo 0
    fi
}

get_current_version()
{
    echo "$TGZ_FILELIST_INSTALLED" | sed -e "s/$LST_EXT$//" | cut -d'-' -f2-3
}

get_package_version()
{
    PKG_FILELIST="`ls -1 $CIMS_INSTALL/$1-[0-9]*$LST_EXT 2> /dev/null`"
    if test -r "$PKG_FILELIST" ; then
        basename "$PKG_FILELIST" | sed -e "s/^$1-//" -e "s/$LST_EXT$//" | cut -d'-' -f1-2
    fi
}

get_installed_dependent_packages()
{
    INSTALLED_PKGS=
    for DEPENDENT_PKG in $DEPENDENT_PKGS ; do
        DEPENDENT_VERSION="`get_package_version $DEPENDENT_PKG`"
        if test -n "$DEPENDENT_VERSION" ; then
            # Installed
            INSTALLED_PKGS="$INSTALLED_PKGS $DEPENDENT_PKG"
        fi
    done
    echo $INSTALLED_PKGS
}

are_all_required_packages_installed()
{
    ALL_INSTALLED=$TRUE
    for REQUIRED_PKG in $REQUIRED_PKGS ; do
        REQUIRED_VERSION="`get_package_version $REQUIRED_PKG`"
        if test -z "$REQUIRED_VERSION" ; then
            # Not installed
            verbose "Required package $REQUIRED_PKG is not installed."
            ALL_INSTALLED=$FALSE
        fi
    done
    return $ALL_INSTALLED
}

is_installed()
{
    TGZ_FILELIST_INSTALLED="`ls -1 $CIMS_INSTALL/$PKG_NAME-[0-9]*$LST_EXT 2> /dev/null`"
    test -r "$TGZ_FILELIST_INSTALLED"
}

install_tgz()
{
    print "Start $ACTING $PKG_NAME $VERSION ..."
    (
        set -e

        test "$DO_SKIP" = $TRUE || install_tgz_pre

        verbose "Extracting $TGZ ..."
        gunzip -c "$TGZ" | (cd / && tar xfp -) 2>> "$LOG_ERROR"

        test "$DO_SKIP" = $TRUE || install_tgz_post

        tar tf "$TGZ" | tac > "$CIMS_INSTALL/$TGZ_FILELIST_NEW" 2>> "$LOG_ERROR"
        cp -pf $PROG_FULL $CIMS_INSTALL/bin/ 2>> "$LOG_ERROR"
        echo ".$CIMS_INSTALL/bin/`basename $PROG_FULL`" >> "$CIMS_INSTALL/$TGZ_FILELIST_NEW" 2>> "$LOG_ERROR"

        if [ "$ACT" = upgrade -a "$VERSION" != "$CURRENT_VERSION" ]; then

            # If the current version (before upgrade) is before 5.5.1,
            # the CentrifyDC package will have all opensource libraries
            # (openssl, openldap and curl). When we check whether a file
            # has been removed in the new package, we need to check the
            # splitted opensource packages as well.
            #
            SPLITTED_PKGS=
            case "$CURRENT_VERSION" in
                5.4.* | 5.5.0*)
                    for PKG in $REQUIRED_PKGS ; do
                        TGZ_FILELIST="$CIMS_INSTALL/$PKG-$VERSION$LST_EXT"
                        if [ -f "$TGZ_FILELIST" ]; then
                            SPLITTED_PKGS="$SPLITTED_PKGS $PKG"
                        else
                            warn "File $TGZ_FILELIST does not exist."
                        fi
                    done
                    verbose "Cleaning will also check with packages:$SPLITTED_PKGS"
                    ;;
            esac

            # Remove files and symlinks that is not in the new package
            verbose "Cleaning old files ..."
            while read -r line ; do
                FOUND=$FALSE
                for PKG in $PKG_NAME $SPLITTED_PKGS ; do
                    TGZ_FILELIST="$CIMS_INSTALL/$PKG-$VERSION$LST_EXT"
                    if grep "^$line$" "$TGZ_FILELIST" > /dev/null ; then
                        FOUND=$TRUE
                        break
                    fi
                done
                if [ "$FOUND" = $FALSE ]; then
                    if [ -f "/$line" -o -h "/$line" ]; then
                        verbose "  deleting /$line"
                        rm -f "/$line" || true
                    fi
                fi
            done < "$TGZ_FILELIST_INSTALLED" 2>> "$LOG_ERROR"

            # Remove directories that is not in the new package
            verbose "Cleaning old directories ..."
            DIR_REMOVED=$TRUE
            while test "$DIR_REMOVED" = $TRUE ; do
                DIR_REMOVED=$FALSE
                while read -r line ; do
                    FOUND=$FALSE
                    for PKG in $PKG_NAME $SPLITTED_PKGS ; do
                        TGZ_FILELIST="$CIMS_INSTALL/$PKG-$VERSION$LST_EXT"
                        if grep "^$line$" "$TGZ_FILELIST" > /dev/null ; then
                            FOUND=$TRUE
                            break
                        fi
                    done
                    if [ "$FOUND" = $FALSE ]; then
                        if [ -d "/$line" -a "/$line" != "/./" ]; then
                            # Directory may belong to system,
                            # do not show error
                            verbose "  deleting /$line"
                            rmdir "/$line" 2> /dev/null && DIR_REMOVED=$TRUE || true
                        fi
                    fi
                done < "$TGZ_FILELIST_INSTALLED" 2>> "$LOG_ERROR"
            done

            rm "$TGZ_FILELIST_INSTALLED" 2>> "$LOG_ERROR" || true
        fi
    )

    RC=$?
    if [ "$ACT" = install -a $RC -ne 0 ]; then
        rm -f "$CIMS_INSTALL/$TGZ_FILELIST_NEW"
    fi

    return $RC
}

uninstall_tgz()
{
    print "Start $ACTING $PKG_NAME $CURRENT_VERSION ..."
    (
        set -e

        if [ ! -r "$TGZ_FILELIST_INSTALLED" ]; then
            error "Cannot read $TGZ_FILELIST_INSTALLED"
            false
        fi

        test "$DO_SKIP" = $TRUE || uninstall_tgz_pre

        # Remove files and symlinks
        verbose "Removing package files ..."
        while read -r line ; do
            if [ -f "/$line" -o -h "/$line" ]; then
                rm -f "/$line" || true
            fi
        done < "$TGZ_FILELIST_INSTALLED" 2>> "$LOG_ERROR"

        # Remove directories
        verbose "Removing package directories ..."
        DIR_REMOVED=$TRUE
        while test "$DIR_REMOVED" = $TRUE ; do
            DIR_REMOVED=$FALSE
            while read -r line ; do
                if [ -d "/$line" -a "/$line" != "/./" ]; then
                    # Directory may belong to system,
                    # do not show error
                    rmdir "/$line" 2> /dev/null && DIR_REMOVED=$TRUE || true
                fi
            done < "$TGZ_FILELIST_INSTALLED" 2>> "$LOG_ERROR"
        done

        test "$DO_SKIP" = $TRUE || uninstall_tgz_post

        rm -f "$TGZ_FILELIST_INSTALLED"
    )
    
    if [ "$DO_FORCE" = $TRUE ] && is_installed ; then
        print "Force $ACTING ..."
        rm -f "$TGZ_FILELIST_INSTALLED"
    fi
    
    test ! -f "$TGZ_FILELIST_INSTALLED"
}

install_tgz_pre()
{
    verbose "Running pre-install script ..."

    # Check if /etc/logrotate.d/centrifydc is a symlink
    LOGROTATE_CONF="/etc/logrotate.d/centrifydc"
    if [ -h "${LOGROTATE_CONF}" ]; then
        echo "ERROR: The file ${LOGROTATE_CONF} is a symbolic link which is not expected." \
            "Cannot proceed with upgrade/installation." \
            "Please contact Centrify Technical Support." | tee -a "$LOG_ERROR"
        exit 1
    fi

    if [ "$ACT" = upgrade -o "$ACT" = reinstall ]; then
        verbose "Pre-install for $ACTING to $VERSION ..."
        CDC_VER_ORIG_FULL=$CURRENT_VERSION

        mkdir -p /var/centrifydc/upgrade

        # Check DirectAudit status
        if [ -x "$CIMS_USR/sbin/dacontrol" ] && [ -x "$CIMS_USR/bin/dainfo" ]; then
            CDA_CUR_VER="`$CIMS_USR/bin/dainfo -v  | cut -d' ' -f3 | cut -d- -f1`"
            log "CDA_CUR_VER=$CDA_CUR_VER"

            if [ `$CIMS_USR/bin/dainfo -s > /dev/null 2>&1; echo $?` -ne 0 ]; then
                verbose "Centrify DirectAudit auditing is currently enabled (symlink hooked)."
                IS_AUDITING=$TRUE
            else
                IS_AUDITING=$FALSE
            fi

            # if auditing is enabled, do nothing and continue
            if [ "$IS_AUDITING" = $FALSE ]; then
                $CIMS_USR/sbin/dacontrol --cdashContWithoutDad >> "$LOG_ERROR" 2>&1
            fi
        fi

        # Should we do cache upgrading at the end ?
        if [ ! -f /var/centrifydc/kset.domain ]; then
            verbose "Computer is not joined, skip cache upgrade."
        elif [ `$CIMS_USR/bin/adinfo -c | grep 'adclient.cache.encrypt:' | grep true > /dev/null; echo $?` -eq 0 ]; then
            warn "Cache is encrypted and it cannot be upgraded."
        elif [ -x "$CIMS_INSTALL/bin/adcache" ]; then
            $CIMS_INSTALL/bin/adcache -s 2>/dev/null | grep 0x > /var/centrifydc/upgrade/CHECK_SUM_ORIG
            CDC_CUR_CHECK_SUM="`cat /var/centrifydc/upgrade/CHECK_SUM_ORIG`"
            verbose "Dictionary checksum: $CDC_CUR_CHECK_SUM"
        fi

        # Remember the original version
        echo "$CDC_VER_ORIG_FULL" > /var/centrifydc/upgrade/CDC_VER_ORIG 2>> "$LOG_ERROR" || true

        # TBD: Remember adnisd, ldapproxy versions and status

        # Stop adclient
        CHECK_ADCLIENT="$CIMS_INSTALL/bin/centrifydc status | grep \"is running\""
        if eval $CHECK_ADCLIENT >> "$LOG_ERROR" 2>&1 ; then
            touch /var/centrifydc/upgrade/START 2>> "$LOG_ERROR" || true
            CDCWATCH_PID=`PROCESS=cdcwatch && eval $GET_PROCESS_PID` || true
            test -n "$CDCWATCH_PID" && kill $CDCWATCH_PID >> "$LOG_ERROR" 2>&1 || true
            
            verbose "Stopping adclient ..."
            $CIMS_INSTALL/bin/centrifydc stop >> "$LOG_ERROR" 2>&1 || true

            while eval $CHECK_ADCLIENT > /dev/null 2> /dev/null ; do
                verbose "  adclient still running"
                sleep 1
            done
            if [ -f /etc/nsswitch.conf.pre_cdc ]; then
                verbose "Backing out /etc/nsswitch.conf ..."
                mv /etc/nsswitch.conf /etc/nsswitch.conf.upgrade 2>> "$LOG_ERROR" || true
                cp /etc/nsswitch.conf.pre_cdc /etc/nsswitch.conf 2>> "$LOG_ERROR" || true
            fi
        else
            rm -f /var/centrifydc/upgrade/START
        fi

        # Remove pid file to avoid cache flush when adclient starts next time
        rm -f /var/run/adclient.pid

        # Stop Centrify-KCM if it's running
        if [ -x "$CIMS_INSTALL/bin/centrify-kcm" ]; then
            if $CIMS_INSTALL/bin/centrify-kcm status | grep "is running" >> "$LOG_ERROR" 2>&1 ; then
                warn "In-memory Kerberos cache credentials will be lost."
                
                verbose "Stopping Centrify-KCM ..."
                $CIMS_INSTALL/bin/centrify-kcm stop >> "$LOG_ERROR" 2>&1 || true
            fi
        fi

        # Backup the reference files that are outside the default directory
        mkdir -p /etc/centrifydc/custom 2>> "$LOG_ERROR" || true
        cp -p /etc/centrifydc/*.* /etc/centrifydc/custom 2>> "$LOG_ERROR" || true
        if $CIMS_INSTALL/bin/upgradeconf -h |grep "\-B" > /dev/null 2>&1 ; then
            $CIMS_INSTALL/bin/upgradeconf \
                -B /etc/centrifydc/custom/centrifydc.conf \
                /etc/centrifydc/custom/rfc.conf >> "$LOG_ERROR" 2>&1 || true
        fi

        # Backup dzshrc
        cp -pf /etc/dzshrc /etc/dzshrc.pre_cdc_upgrade 2>> "$LOG_ERROR" || true
    fi
}

install_tgz_post()
{
    verbose "Running post-install script ..."
    ### Start setup the environment that will not be reverted in uninstall

    # Inject environment
    verbose "Checking global environment ..."
    update_file /etc/environment \
        "$INJECT_ENV" 2>> "$LOG_ERROR"

    # Inject systemd environment
    verbose "Checking systemd system environment ..."
    update_file /etc/systemd/system.conf \
        "DefaultEnvironment=$INJECT_ENV" 2>> "$LOG_ERROR"
    if [ "$UPDATE_RESULT" = updated ]; then
        # Restart of systemd is required
        verbose "Restart of systemd is required."
    fi

    # Enable nscd service
    verbose "Checking nscd service ..."
    if systemctl list-unit-files | grep "nscd\.service" > /dev/null ; then

        verbose "Checking nscd configuration ..."
        symlink_to_file /etc/nscd.conf 2>> "$LOG_ERROR"

        if systemctl is-enabled nscd.service > /dev/null ; then
            verbose "The nscd service is already enabled."
        else
            verbose "Enabling nscd service ..."
            systemctl enable nscd.service 2>> "$LOG_ERROR" || true
        fi
        if [ "`systemctl --property=ActiveState show nscd.service`" != \
            "ActiveState=active" ]; then
            verbose "Startng nscd service ..."
            systemctl start nscd.service 2>> "$LOG_ERROR" || true
        fi
    else
        warn "No nscd service is installed on the system."
    fi

    # Configure NSS
    verbose "Checking NSS configuration ..."
    symlink_to_file /etc/nsswitch.conf 2>> "$LOG_ERROR"
    
    # Configure PAM
    #
    # There is no /usr/lib/pam.d on Atomic Host. Instead, all
    # PAM files are already in the /etc/pam.d directory. Besides,
    # it uses system-auth-ac file and system-auth is a symlink to
    # that file.
    #
    # That means, we should not do anything on Atomic Host.
    #
    verbose "Checking PAM configuration ..."
    if [ -d "$PAM_USR_DIR" ]; then
        mkdir -p "$PAM_ETC_DIR"
        symlink_to_file "$PAM_ETC_DIR/system-auth" 2>> "$LOG_ERROR"
        copy_pam_file system-auth 2>> "$LOG_ERROR"

        # The login file is for dzdo PAM file installation later on
        symlink_to_file "$PAM_ETC_DIR/login" 2>> "$LOG_ERROR"
        copy_pam_file login 2>> "$LOG_ERROR"
    fi

    # Configure sshd
    verbose "Checking sshd configuration ..."
    update_file /etc/ssh/sshd_config "ChallengeResponseAuthentication yes" \
        "^[ 	]*ChallengeResponseAuthentication[ 	][ 	]*yes[ 	]*$" \
        2>> "$LOG_ERROR"

    ### End setup

    # Add dzsh
    symlink_to_file "/etc/shells" 2>> "$LOG_ERROR"
    if [ -f /etc/shells ]; then
        if ! grep "$CIMS_USR/bin/dzsh" /etc/shells > /dev/null ; then
            verbose "Adding $CIMS_USR/bin/dzsh to /etc/shells ..."
            echo "$CIMS_USR/bin/dzsh" >> /etc/shells
        fi
    fi

    # Copy the right dzdo pam config to /etc/pam.d 
    if [ -d "/etc/pam.d" ] ; then
        PAM_CONF_SUFFIX=
        if [ -e "/etc/pam.d/common-auth" ] ; then 
            if grep "@include" /etc/pam.d/login > /dev/null; then
                PAM_CONF_SUFFIX=-common-debian
            else
                if grep "pam_umask.so" /etc/pam.d/common-session > /dev/null; then
                    PAM_CONF_SUFFIX=-common-suse
                else
                    PAM_CONF_SUFFIX=-common
                fi
            fi
        elif [ -e "/etc/pam.d/system-auth" ] ; then 
            if grep "pam_stack.so" /etc/pam.d/login > /dev/null; then
                PAM_CONF_SUFFIX=-stack
            else
                PAM_CONF_SUFFIX=-system
            fi
        elif grep "pam_unix.so" /etc/pam.d/login > /dev/null; then
            PAM_CONF_SUFFIX=-normal
        elif grep "pam_unix2.so" /etc/pam.d/login > /dev/null; then
            PAM_CONF_SUFFIX=-normal2
        elif [ -e "/etc/pam.d/sudo" ] ; then 
            verbose "Installing dzdo PAM config file from /etc/pam.d/sudo ..."
            cp -p /etc/pam.d/sudo /etc/pam.d/dzdo 2>> "$LOG_ERROR"
        else
            warn "Can't set /etc/pam.d/dzdo successfully. Please contact administrator to create it manually."
        fi
        if [ -n "$PAM_CONF_SUFFIX" ]; then
            verbose "Installing dzdo PAM config files from *-pam-conf$PAM_CONF_SUFFIX ..."
            for FILE in $CIMS_INSTALL/etc/*-pam-conf$PAM_CONF_SUFFIX ; do
                PAM_NAME=`basename $FILE | sed -e 's/-pam-conf.*//'`
                verbose "  /etc/pam.d/$PAM_NAME"
                cp -p "$FILE" "/etc/pam.d/$PAM_NAME" 2>> "$LOG_ERROR"
            done
        fi
    fi

    # Create necessary files and symlinks
    verbose "Checking certs directory ..."
    mkdir -p /var/centrify/net
    if [ -d "$SSL_CERTS_DIR" ]; then
        if [ ! -e /var/centrify/net/certs ]; then
            verbose "Creating symlink /var/centrify/net/certs ..."
            ln -s "$SSL_CERTS_DIR" \
                /var/centrify/net/certs 2>> "$LOG_ERROR" || true
        else
            verbose "/var/centrify/net/certs already exists."
        fi
    else
        warn "The certs directory $SSL_CERTS_DIR does not exist."
    fi

    # Install systemd services
    verbose "Installing systemd services ..."
    mkdir -p "$SYSTEMD_DIR"
    install_systemd_service centrifydc.service 2>> "$LOG_ERROR"
    install_systemd_service centrify-kcm.service 2>> "$LOG_ERROR"

    if [ "$ACT" = upgrade -o "$ACT" = reinstall ]; then
        verbose "Post-install for $ACTING to $VERSION ..."

        # Save new files
        mkdir -p /etc/centrifydc/new 2>> "$LOG_ERROR" || true
        cp -pf /etc/centrifydc/*.* /etc/centrifydc/new/. 2>> "$LOG_ERROR" || true

        # Restore customized files
        cp -pf /etc/centrifydc/custom/*.* /etc/centrifydc/. 2>> "$LOG_ERROR" || true

        # Output directory
        rm -rf /etc/centrifydc/merge 2>> "$LOG_ERROR"
        mkdir -p /etc/centrifydc/merge 2>> "$LOG_ERROR" || true

        # Original CDC version
        CDC_VER_ORIG=
        if [ -s /var/centrifydc/upgrade/CDC_VER_ORIG ]; then
            CDC_VER_ORIG="`cat /var/centrifydc/upgrade/CDC_VER_ORIG | sed 's/[^0-9]*//g'`" || true
        fi

        # TBD: auto.schema.name.lower

        verbose "Upgrading configuration files ..."
        CUSTOMCDCCONF="/etc/centrifydc/custom/centrifydc.conf"
        CUSTOMCDCCONFTMP="${CUSTOMCDCCONF}.tmp"
        # Remove $CIMS_USR/sbin/daflush from cli.hook.adflush if it exists
        # Note the legal assignment tokens are: " ", " =",  ":", ":="
        # Copy to preserve file ownership and permissions
        cp -pf ${CUSTOMCDCCONF} ${CUSTOMCDCCONFTMP} 2>> "$LOG_ERROR" || true
        sed '/^[ 	]*cli.hook.adflush[ :]=\{0,1\}[ 	]*\@INSTALL_USR_DIR@\/sbin\/daflush[ 	]*.*/d' ${CUSTOMCDCCONF} > ${CUSTOMCDCCONFTMP} 2>> "$LOG_ERROR" || true
        mv -f ${CUSTOMCDCCONFTMP} ${CUSTOMCDCCONF} 2>> "$LOG_ERROR"

        RC=0
        if [ -f /etc/centrifydc/custom/rfc.conf ]; then
            $CIMS_INSTALL/bin/upgradeconf \
                -r -c /etc/centrifydc/new/upgradeconf.conf \
                -R /etc/centrifydc/custom/rfc.conf \
                /etc/centrifydc/custom/centrifydc.conf \
                /etc/centrifydc/new/centrifydc.conf \
                /etc/centrifydc/merge/centrifydc.conf \
                >> "$LOG_ERROR" 2>&1 || RC=$?
        else
            $CIMS_INSTALL/bin/upgradeconf \
                -r -c /etc/centrifydc/new/upgradeconf.conf \
                /etc/centrifydc/custom/centrifydc.conf \
                /etc/centrifydc/new/centrifydc.conf \
                /etc/centrifydc/merge/centrifydc.conf \
                >> "$LOG_ERROR" 2>&1 || RC=$?
        fi

        # If upgrade successfully, update the config files
        if [ $RC -eq 0 ] ; then
            verbose "Replacing configuration files ..."
            for FILE in /etc/centrifydc/merge/*.* ; do
                if [ -f "$FILE" -a -s "$FILE" ]; then
                    verbose "  moving $FILE to /etc/centrifydc/"
                    mv -f "$FILE" /etc/centrifydc/. 2>> "$LOG_ERROR" || true
                fi
            done
            rm -rf /etc/centrifydc/merge
        else
            warn "Failed to merge configuration files."
        fi

        # Old cdc may leave apu.lst file there
        rm -f /etc/centrifydc/apu.lst
        
        # If the dzshrc changed, use the customized one.
        if [ -f /etc/dzshrc -a -f /etc/dzshrc.pre_cdc_upgrade ]; then
            if diff /etc/dzshrc /etc/dzshrc.pre_cdc_upgrade > /dev/null 2>&1 ; then
                verbose "No changes found in /etc/dzshrc"
                rm -f /etc/dzshrc.pre_cdc_upgrade
            else
                verbose "Preserving /etc/dzshrc ..."
                mv -f /etc/dzshrc /etc/dzshrc.cdc 2>> "$LOG_ERROR" && \
                    mv -f /etc/dzshrc.pre_cdc_upgrade /etc/dzshrc 2>> "$LOG_ERROR" || true
            fi
        fi
        
        # TBD: Run virtual registry migration tool

        verbose "Executing fix_pam_rsh.sh ..."
        $CIMS_INSTALL/bin/fix_pam_rsh.sh >> "$LOG_ERROR" 2>&1 || true

        # Check whether this service is enabled before upgrade
        if [ "$SYSTEMD_IS_ENABLED" = "$TRUE" ]; then
            # enable service for startup
            systemctl enable centrifydc.service >> "$LOG_ERROR" 2>&1 || true
        fi

        # Do cache upgrade
        if [ -f /var/centrifydc/tmp/CENTRIFY_FORCE_CACHE_DELETE ]; then
            verbose "Found /var/centrifydc/tmp/CENTRIFY_FORCE_CACHE_DELETE, removing cache ..."
            ls -l /var/centrifydc/*.cache /var/centrifydc/*.idx >> "$LOG_ERROR" 2> /dev/null || true
            rm -f /var/centrifydc/*.cache /var/centrifydc/*.idx 2>> "$LOG_ERROR"
            rm -f /var/centrifydc/tmp/CENTRIFY_FORCE_CACHE_DELETE 2>> "$LOG_ERROR"
        elif [ ! -f /var/centrifydc/kset.domain ]; then
            verbose "Computer is not joined, skip cache upgrade."
        elif [ `$CIMS_USR/bin/adinfo -c | grep 'adclient.cache.encrypt:' | grep true > /dev/null; echo $?` -eq 0 ]; then
            warn "Cache is encrypted and it cannot be upgraded."
            if ls -l /var/centrifydc/*.cache >> "$LOG_ERROR" 2> /dev/null ; then
                verbose "Removing *.cache files ..."
                rm -f /var/centrifydc/*.cache 2>> "$LOG_ERROR"
            fi
            if ls -l /var/centrifydc/*.idx >> "$LOG_ERROR" 2> /dev/null ; then
                verbose "Removing *.idx files ..."
                rm -f /var/centrifydc/*.idx 2>> "$LOG_ERROR"
            fi
        elif [ -s /var/centrifydc/upgrade/CHECK_SUM_ORIG ]; then
            CDC_CUR_CHECK_SUM="`cat /var/centrifydc/upgrade/CHECK_SUM_ORIG`" || true
            verbose "Upgrading cache from checksum ${CDC_CUR_CHECK_SUM} ..."
            if ! $CIMS_INSTALL/bin/adcache -F "${CDC_CUR_CHECK_SUM}" >> "$LOG_ERROR" 2>&1 ; then
                warn "Cache upgrade exited with error, please check log file for details."
            fi
        elif [ -s /var/centrifydc/upgrade/CDC_VER_ORIG ]; then
            CDC_CUR_VER_FULL="`cat /var/centrifydc/upgrade/CDC_VER_ORIG`" || true
            verbose "Upgrading cache from version ${CDC_CUR_VER_FULL} ..."
            if ! $CIMS_INSTALL/bin/adcache -F "${CDC_CUR_VER_FULL}" >> "$LOG_ERROR" 2>&1 ; then
                warn "Cache upgrade exited with error, please check log file for details."
            fi
        fi

        # Start adclient
        if [ -f /var/centrifydc/upgrade/START ]; then
            if [ -f "/etc/nsswitch.conf.upgrade" ]; then
                verbose "Restoring /etc/nsswitch.conf ..."
                mv /etc/nsswitch.conf.upgrade /etc/nsswitch.conf 2>> "$LOG_ERROR" || true
            fi
            verbose "Starting adclient ..."
            $CIMS_INSTALL/bin/centrifydc start >> "$LOG_ERROR" 2>&1 || true

            # Update keytab if adclient is in connected status
            verbose "Start to update keytab ..."
            ADCLIENT_MODE=""
            for i in 1 2 3 4 ; do
                ADCLIENT_MODE="`$CIMS_USR/bin/adinfo -m`" || true
                if [ "X${ADCLIENT_MODE}" = "Xstarting" -o "X${ADCLIENT_MODE}" = "X<unavailable>" ]; then
                    verbose "  adclient is ${ADCLIENT_MODE}, wait for $i second(s)."
                    sleep $i
                else
                    break
                fi
            done
            
            if [ "X${ADCLIENT_MODE}" = "Xconnected" ]; then
                verbose "Updating keytab ..."
                $CIMS_USR/sbin/adkeytab -C -m >> "$LOG_ERROR" 2>&1 || true
            else
                verbose "Skip to update keytab. Adclient is ${ADCLIENT_MODE}."
            fi

            rm /var/centrifydc/upgrade/START >> "$LOG_ERROR" 2>&1 || true
        fi

        # TBD: Restart other Centrify daemons, e.g. adnisd, ldapproxy, samba

        # Restart dad
        verbose "Checking dad ..."
        PID=`PROCESS=dad && eval $GET_PROCESS_PID` || true
        if [ ! -z "$PID" ]; then
            verbose "Restarting dad ..."
            ps -ae | awk '$4 ~ /^dad$/' >> "$LOG_ERROR" 2>&1 || true
            for tries in 1 2 3 4 5 ; do
                dadPid=`PROCESS=dad && eval $GET_PROCESS_PID` || true
                if [ ! -z "$dadPid" ]; then
                    case "$tries" in
                        1)
                            verbose "Stopping dad by centrifyda stop ..."
                            if [ -x $CIMS_INSTALL/bin/centrifyda ]; then
                                $CIMS_INSTALL/bin/centrifyda stop >> "$LOG_ERROR" 2>&1 || true
                            elif [ -x /etc/init.d/centrifyda ]; then
                                /etc/init.d/centrifyda stop >> "$LOG_ERROR" 2>&1 || true
                            fi
                            ;;
                        2)
                            verbose "Stopping dad by dastop ..."
                            $CIMS_USR/sbin/dastop >> $"LOG_ERROR" 2>&1 || true
                            ;;
                        3 | 4)
                            verbose "Stopping dad by kill ..."
                            kill $dadPid >> "$LOG_ERROR" 2>&1 || true
                            ;;
                        5)
                            verbose "Stopping dad by kill ..."
                            kill -9 $dadPid >> "$LOG_ERROR" 2>&1 || true
                            ;;
                    esac
                    sleep 5
                else
                    verbose "DirectAudit daemon status: not running ..."
                    break
                fi
            done
            if [ -x $CIMS_INSTALL/bin/centrifyda ]; then
                $CIMS_INSTALL/bin/centrifyda start >> "$LOG_ERROR" 2>&1 || true
            elif [ -x /etc/init.d/centrifyda ]; then
                /etc/init.d/centrifyda start >> "$LOG_ERROR" 2>&1 || true
            fi
            ps -ae | awk '$4 ~ /^dad$/' >> "$LOG_ERROR" 2>&1 || true
        fi

        if [ -x "$CIMS_USR/sbin/dacontrol" ] && [ -x "$CIMS_USR/bin/dainfo" ]; then
            if $CIMS_USR/bin/dainfo > /dev/null 2>&1 ; then
                CDA_CUR_VER="`$CIMS_USR/bin/dainfo -v  | cut -d' ' -f3 | cut -d- -f1`"
                log "CDA_CUR_VER=${CDA_CUR_VER}"

                $CIMS_USR/sbin/dacontrol --cdashHangWithoutDad >> "$LOG_ERROR" 2>&1 || true
            fi
        fi

    else
        # TBD: Enable centrifydc service
        verbose "Enabling Centrify DirectControl service ..."
        systemctl daemon-reload 2>> "$LOG_ERROR" || true
        systemctl enable centrifydc.service 2>> "$LOG_ERROR" || true

        # Set license mode
        verbose "Setting license mode ..."
        adlicense -l >> "$LOG_ERROR" 2>> "$LOG_ERROR"
    fi
}

uninstall_tgz_pre()
{
    verbose "Running pre-uninstall script ..."
    # Turn addebug off
    if addebug | grep off >> "$LOG_ERROR" 2>&1 ; then
        :
    else
        verbose "Disabling addebug ..."
        if ! addebug off >> "$LOG_ERROR" 2>&1 ; then
            warn "Cannot disable addebug"
        fi
    fi

    # Stop and disable centrifydc service
    systemctl stop centrify-kcm.service 2>> "$LOG_ERROR" || true
    systemctl stop centrifydc.service 2>> "$LOG_ERROR" || true
    systemctl disable centrify-kcm.service 2>> "$LOG_ERROR" || true
    systemctl disable centrifydc.service 2>> "$LOG_ERROR" || true

    # Uninstall systemd services
    verbose "Removing systemd services ..."
    rm -f "$SYSTEMD_DIR/centrify-kcm.service"
    rm -f "$SYSTEMD_DIR/centrifydc.service"

    # Remove files and symlinks
    rm -f /var/centrify/net/certs

    # Remove PAM config files that we create
    rm -f /etc/pam.d/dzdo
    rm -f /etc/pam.d/dzdo-*

    if [ -f /etc/shells ]; then
        verbose "Removing $CIMS_USR/bin/dzsh from /etc/shells ..."
        cp -pf /etc/shells /etc/shells.tmp 2>> "$LOG_ERROR"
        sed "\:$CIMS_USR/bin/dzsh:d" /etc/shells > /etc/shells.tmp 2>> "$LOG_ERROR"
        mv -f /etc/shells.tmp /etc/shells 2>> "$LOG_ERROR"
        if awk '/^\//{ exit 1 }' /etc/shells ; then
            warn "/etc/shells is empty, all ftp access will be disabled."
        fi
    fi

    # Remove configuration files that may exist
    rm -f /etc/centrifydc/passwd.ovr
    rm -f /etc/centrifydc/group.ovr
    rm -f /etc/centrifydc/openldap/slapd.conf.prev
    rm -f /etc/centrifydc/prev_log_setting_off
    rm -f /etc/centrifydc/share/apu.lst
}

uninstall_tgz_post()
{
    verbose "Running post-uninstall script ..."
    # Remove all files left on the system
    if [ -d /var/centrify/cloud ]; then
        for f in /var/centrify/* ; do
            if [ "$f" != "/var/centrify/cloud" ]; then
                verbose "Removing $f ..."
                rm -rf $f > /dev/null
            fi
        done
    else
        remove_directory /var/centrify > /dev/null || true
    fi
    remove_directory /var/centrifydc > /dev/null || true
    remove_directory /etc/centrifydc > /dev/null || true
    remove_directory /tmp/precreate > /dev/null || true
    #kset.* by non-root users
    rm -f /tmp/kset.* > /dev/null || true
}

clean_up()
{
    rm -f "$TMP_FILE"
}

clean_exit()
{
    clean_up
    exit $1
}

signal_handler()
{
    trap "" ${SIGNAL_LIST}
    error "Signal received."
    clean_up
    exit 1
}


# Parse arguments
while getopts :ieurfsph OPT; do
    case $OPT in
        i)
            ACT=install
            ;;
        e)
            ACT=uninstall
            ;;
        u)
            ACT=upgrade
            ;;
        r)
            ACT=reinstall
            ;;
        f)
            DO_FORCE=$TRUE
            ;;
        s)
            DO_SKIP=$TRUE
            ;;
        p)
            DO_VERBOSE=$TRUE
            ;;
        h)
            usage
            exit 0
            ;;
        \?)
            error "Invalid option '-$OPTARG'."
            exit 1
            ;;
        :)
            error "'-$OPTARG' requires an argument."
            exit 1
            ;;
    esac
done
shift $((OPTIND-1))

if [ -z "$ACT" ]; then
    error "No operation is specified."
    usage
    exit 1
fi

case "$ACT" in
    install|upgrade|reinstall|uninstall)
        if [ "`id -u`" != 0 ]; then
            error "This script must be run by root."
            exit 1
        fi
        ;;
esac

# Write a marker to the log file
DO_LOG=$TRUE
LOG_ERROR=$LOG_FILE
log ""
log "--- $PROG.sh ---:" `date`
log "Command line: $COMMAND_LINE"

trap signal_handler ${SIGNAL_LIST}
touch "$TMP_FILE" 2>> "$LOG_ERROR"

# Preprocess for install, upgrade and reinstall
case "$ACT" in
    install|upgrade|reinstall)
        if [ -f "$TGZ" ]; then
            if [ "$ACT" = upgrade -o "$ACT" = reinstall ] && ! is_installed ; then
                ACT=install
            fi
            if [ "$ACT" = install ]; then
                ACTED=installed
                ACTING=installing
                if is_installed ; then
                    error "$PKG_NAME is already installed."
                    clean_exit 1                    
                fi
            else
                VERSION_NUM=`get_version_number "$VERSION"`
                CURRENT_VERSION=`get_current_version`
                CURRENT_VERSION_NUM=`get_version_number "$CURRENT_VERSION"`
                if [ -n "$CURRENT_VERSION" ]; then
                    if [ "$ACT" = upgrade ]; then
                        ACTED=upgraded
                        ACTING=upgrading
                        if [ $VERSION_NUM -lt $CURRENT_VERSION_NUM ]; then
                            error "$PKG_NAME installed version $CURRENT_VERSION is newer than $VERSION."
                            clean_exit 1
                        fi
                    else
                        ACTED=reinstalled
                        ACTING=reinstalling
                        if [ "$CURRENT_VERSION" != "$VERSION" ]; then
                            error "$PKG_NAME installed version is not $VERSION."
                            clean_exit 1
                        fi
                    fi
                else
                    error "$PKG_NAME version cannot be determined."
                    clean_exit 1
                fi
            fi
            if ! are_all_required_packages_installed ; then
                error "Please install all required packages first!"
                clean_exit 1
            fi
        else
            error "$PKG_NAME tarball not found."
            clean_exit 1
        fi
        ;;
esac

case "$ACT" in
    install|upgrade|reinstall)

        install_tgz
        RC=$?
        if [ $RC -eq 0 ]; then
            print "$PKG_NAME is successfully $ACTED."
            
            if [ ! "$SYSTEMD_CONF" -ot "$TMP_FILE" ]; then
                print "Please reboot the system to ensure all services use the new settings."
            elif [ ! /etc/ssh/sshd_config -ot "$TMP_FILE" ]; then
                print "Please restart sshd to ensure sshd uses the new settings."
            fi
        else
            error "Cannot $ACT $PKG_NAME."
            clean_exit 1
        fi
        ;;
    uninstall)
        ACTED=uninstalled
        ACTING=uninstalling
        if is_installed ; then

            CURRENT_VERSION=`get_current_version`
            CURRENT_VERSION_NUM=`get_version_number "$CURRENT_VERSION"`
            DEPENDENT_PKGS=`get_installed_dependent_packages`
            if [ -n "$DEPENDENT_PKGS" ]; then
                error "Cannot $ACT because ${DEPENDENT_PKGS} are currently installed."
                clean_exit 1
            fi

            # Check if machine is joined
            if [ -s /var/centrifydc/kset.domain ]; then
                error "Cannot $ACT while the machine is joined."
                clean_exit 1
            fi

            uninstall_tgz
            RC=$?
            if [ $RC -eq 0 ]; then
                print "$PKG_NAME is successfully $ACTED."
            else
                error "Cannot $ACT $PKG_NAME."
                clean_exit 1
            fi
        else
            error "$PKG_NAME is not installed."
            clean_exit 1
        fi
        ;;
    *)
        error "Unknown operation."
        clean_exit 1
esac

clean_up
exit 0
